"""
TODO: This has to use the 'yta_web_scraper' library and
also must be in another library, not this one.

I comment it to avoid errors.
"""
# from selenium.webdriver.common.by import By
# from selenium.webdriver.support import expected_conditions as EC
# from selenium.webdriver.support.ui import WebDriverWait
# from selenium.webdriver.common.action_chains import ActionChains
# from selenium.webdriver.common.keys import Keys
# from yta_general_utils.web.scraper.chrome_scraper import go_to_and_wait_loaded, start_chrome

# import time
# import os
# import pyautogui
# import glob
# import shutil

# MESSAGE_SENDER_ME = 1
# MESSAGE_SENDER_OTHER = 2
# PROJECT_ABSOLUTE_PATH = os.getenv('PROJECT_ABSOLUTE_PATH')
# TOUSE_ABSOLUTE_PATH = os.getenv('TOUSE_ABSOLUTE_PATH')
# # TODO: Move this to .env variable
# DOWNLOADS_ABSOLUTE_PATH = 'C:/Users/dania/Downloads/'

# def __setup(driver, name, conversation_date, profile_picture_absolute_path):
#     # TODO: Set maximum 'name' and 'username' lengths
#     # TODO: Implement 'status' (?)
#     __set_name(driver, name)
#     __set_status(driver, 'En línea')
#     __set_conversation_date(driver, conversation_date)
#     __set_profile_picture(driver, profile_picture_absolute_path)

# def __set_profile_picture(driver, absolute_path):
#     element_present = EC.presence_of_element_located((By.ID, 'profileImg'))
#     element = WebDriverWait(driver, 10).until(element_present)
#     time.sleep(1)
#     element = element.find_element(By.XPATH, '..')
#     element.click()
#     time.sleep(2)
#     pyautogui.write((absolute_path).replace('/', '\\')) 
#     time.sleep(1)
#     pyautogui.press('enter')
#     time.sleep(1)

# def __set_conversation_date(driver, str):
#     date_input = driver.find_element(By.ID, 'edit-chatday')
#     time.sleep(1)
#     date_input.clear()
#     date_input.send_keys(str)

# def __send_my_photo(driver, photo_absolute_path, hour):
#     driver.execute_script('window.scrollTo(0, 0)')
#     profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person2")]')
#     profile2_tab.click()
#     element_present = EC.presence_of_element_located((By.ID, 'SecondImage'))
#     element = WebDriverWait(driver, 10).until(element_present)
#     time.sleep(1)
#     element = element.find_element(By.XPATH, '..')
#     element.click()
#     time.sleep(2)
#     pyautogui.write((photo_absolute_path).replace('/', '\\')) 
#     time.sleep(1)
#     pyautogui.press('enter')
#     time.sleep(1)
#     profile2_time_input = driver.find_element(By.ID, 'time-profile2')
#     profile2_time_input.clear()
#     profile2_time_input.send_keys(hour)
#     time.sleep(1)
#     # Set message as read
#     profile2_message_is_read_radiobutton = driver.find_element(By.ID, 'read')
#     if not profile2_message_is_read_radiobutton.is_selected():
#         profile2_message_is_read_radiobutton.click()
#     time.sleep(1)
#     profile2_textarea = driver.find_element(By.ID, 'profile2_message')
#     profile2_textarea.clear()
#     profile2_textarea.send_keys('')
#     time.sleep(1)
#     actions = ActionChains(driver)
#     for i in range(2):
#         actions.send_keys(Keys.TAB)
#     actions.send_keys(Keys.SPACE)
#     actions.perform()
#     time.sleep(1)

# def __send_other_photo(driver, photo_absolute_path, hour):
#     driver.execute_script('window.scrollTo(0, 0)')
#     profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person1")]')
#     profile2_tab.click()
#     element_present = EC.presence_of_element_located((By.ID, 'InputFirstImage'))
#     element = WebDriverWait(driver, 10).until(element_present)
#     time.sleep(1)
#     element = element.find_element(By.XPATH, '..')
#     element.click()
#     time.sleep(2)
#     pyautogui.write((photo_absolute_path).replace('/', '\\')) 
#     time.sleep(1)
#     pyautogui.press('enter')
#     time.sleep(1)
#     profile1_time_input = driver.find_element(By.ID, 'time-profile1')
#     profile1_time_input.clear()
#     profile1_time_input.send_keys(hour)
#     time.sleep(1)
#     profile1_textarea = driver.find_element(By.ID, 'profile1_message')
#     profile1_textarea.clear()
#     profile1_textarea.send_keys('')
#     time.sleep(1)
#     actions = ActionChains(driver)
#     actions.send_keys(Keys.TAB)
#     actions.send_keys(Keys.SPACE)
#     actions.perform()
#     time.sleep(1)

# def __set_name(driver, name):
#     profile1_name_input = driver.find_element(By.ID, 'profile1_name')
#     time.sleep(1)
#     profile1_name_input.clear()
#     profile1_name_input.send_keys(name)

# def __set_status(driver, status):
#     status_input = driver.find_element(By.ID, 'edit-Status')
#     time.sleep(1)
#     status_input.clear()
#     status_input.send_keys(status)

# def __download(driver, output_filename):
#     time.sleep(2)
#     download_button = driver.find_element(By.ID, 'snapshot')
#     download_button.click()
#     time.sleep(10)
#     # Move downloaded file to our destination
#     list_of_files = glob.glob(DOWNLOADS_ABSOLUTE_PATH + '*.png')
#     latest_file = max(list_of_files, key = os.path.getctime)
#     shutil.move(latest_file, os.path.join(PROJECT_ABSOLUTE_PATH + output_filename))

# def __write_my_message(driver, message, hour):
#     driver.execute_script('window.scrollTo(0, 0)')
#     profile2_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person2")]')
#     profile2_tab.click()
#     time.sleep(1)
#     profile2_time_input = driver.find_element(By.ID, 'time-profile2')
#     profile2_time_input.clear()
#     profile2_time_input.send_keys(hour)
#     time.sleep(1)
#     # Set message as read
#     profile2_message_is_read_radiobutton = driver.find_element(By.ID, 'read')
#     if not profile2_message_is_read_radiobutton.is_selected():
#         profile2_message_is_read_radiobutton.click()
#     time.sleep(1)
#     profile2_textarea = driver.find_element(By.ID, 'profile2_message')
#     profile2_textarea.clear()
#     profile2_textarea.send_keys(message)
#     time.sleep(1)
#     actions = ActionChains(driver)
#     for i in range(2):
#         actions.send_keys(Keys.TAB)
#     actions.send_keys(Keys.SPACE)
#     actions.perform()
#     time.sleep(1)

# def __write_other_message(driver, message, hour):
#     driver.execute_script('window.scrollTo(0, 0)')
#     profile1_tab = driver.find_element(By.XPATH, '//a[contains(@href, "#tab-person1")]')
#     profile1_tab.click()
#     time.sleep(1)
#     profile1_time_input = driver.find_element(By.ID, 'time-profile1')
#     profile1_time_input.clear()
#     profile1_time_input.send_keys(hour)
#     time.sleep(1)
#     profile1_textarea = driver.find_element(By.ID, 'profile1_message')
#     profile1_textarea.clear()
#     profile1_textarea.send_keys(message)
#     time.sleep(1)
#     actions = ActionChains(driver)
#     actions.send_keys(Keys.TAB)
#     actions.send_keys(Keys.SPACE)
#     actions.perform()
#     time.sleep(1)

# # TODO: Implement sending foto
    
# def generate_conversation(name, messages, conversation_date, profile_picture_absolute_path, output_filename):
#     # Messages must come with autor and text
#     URL = 'https://prankshit.com/fake-whatsapp-chat-generator.php'

#     profile_picture_absolute_path = profile_picture_absolute_path.replace('/', '\\')

#     try:
#         driver = start_chrome(True)
#         go_to_and_wait_loaded(driver, URL)

#         # I force 'conversation_date' by now
#         conversation_date = 'Hoy'

#         time.sleep(2)
#         __setup(driver, name, conversation_date, profile_picture_absolute_path)
#         time.sleep(1)

#         # Add more data ('active')
#         messages = [
#             {
#                 'sender': MESSAGE_SENDER_ME,
#                 'message': 'Holaaaaa!',
#                 'hour': '22:07',
#             },
#             {
#                 'sender': MESSAGE_SENDER_ME,
#                 'message': '',
#                 'photo': 'C:/Users/dania/Desktop/PROYECTOS/Youtube/stockpediayt/software/ai.png',
#                 'hour': '22:07',
#             },
#             {
#                 'sender': MESSAGE_SENDER_ME,
#                 'message': 'estás??',
#                 'hour': '22:07',
#             },
#             {
#                 'sender': MESSAGE_SENDER_OTHER,
#                 'message': 'Sí, ¿qué pasa wacho??',
#                 'hour': '22:09',
#             },
#             {
#                 'sender': MESSAGE_SENDER_OTHER,
#                 'message': '',
#                 'photo': 'C:/Users/dania/Desktop/PROYECTOS/Youtube/stockpediayt/software/ai_image.png',
#                 'hour': '22:10',
#             },
#             {
#                 'sender': MESSAGE_SENDER_ME,
#                 'message': 'Nada, solo quería saludarte!! Campeón!',
#                 'hour': '22:10',
#             },
#             {
#                 'sender': MESSAGE_SENDER_OTHER,
#                 'message': 'Dalee! Grande, gracias!!',
#                 'hour': '22:11'
#             },
#         ]

#         for index, message in enumerate(messages):
#             if message['sender'] == MESSAGE_SENDER_ME:
#                 if 'photo' in message and message['photo']:
#                     __send_my_photo(driver, message['photo'], message['hour'])
#                 else:
#                     __write_my_message(driver, message['message'], message['hour'])
#             else:
#                 if 'photo' in message and message['photo']:
#                     __send_other_photo(driver, message['photo'], message['hour'])
#                 else:
#                     __write_other_message(driver, message['message'], message['hour'])

#             __download(driver, 'wip/test_whatsapp_' + str(index) + '.png')
#     finally:
#         driver.close()


from yta_validation.parameter import ParameterValidator


class ConsolePrinter:
    """
    Class to wrap functionality related to printing
    messages in the console.
    """

    @staticmethod
    def print_error(
        message: str = ''
    ):
        ParameterValidator.validate_string('message', message, do_accept_empty = True)

        print(f'>>>> [ERROR] <<<< {message}')

    @staticmethod
    def print_in_progress(
        message: str = ''
    ):
        ParameterValidator.validate_string('message', message, do_accept_empty = True)

        print(f'.... {message}')

    @staticmethod
    def print_completed(
        message: str = ''
    ):
        ParameterValidator.validate_string('message', message, do_accept_empty = True)
        
        print(f'[OK] {message}')


from yta_general_utils.url.dataclasses import UrlParameter, UrlParameters
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union


class UrlHandler:
    """
    Class to wrap the functionality related with urls.
    """

    @staticmethod
    def encode_url_parameter(
        parameter: Union[UrlParameter, str]
    ) -> UrlParameter:
        """
        Get the provided 'parameter' encoded based on
        the RFC 3986.

        - From 'hola dani' to 'hola%20dani'.
        - From 'key=value largo' to 'key%3Dvalue%20largo'
        """
        ParameterValidator.validate_mandatory_instance_of('parameter', parameter, [UrlParameter, str])

        return (
            UrlParameter.from_str(parameter)
            if PythonValidator.is_string(parameter) else
            parameter 
        ).encoded
    
    @staticmethod
    def encode_url_parameters(
        parameters: Union[UrlParameters, dict]
    ) -> UrlParameters:
        """
        Get the provided 'parameters' encoded based on
        the RFC 3986.

        - From 'hola dani&do_force' to 'hola%20dani&do_force'.
        - From 'key=value largo&do_force' to 'key%3Dvalue%20largo&do_forceo'
        """
        ParameterValidator.validate_mandatory_instance_of('parameters', parameters, [UrlParameters, dict])

        return (
            UrlParameters.from_dict(parameters)
            if PythonValidator.is_dict(parameters) else
            parameters
        ).encoded


from yta_validation.parameter import ParameterValidator
from yta_constants.regex import GeneralRegularExpression

import requests
import mimetypes


class UrlValidator:
    """
    Class to wrap functionality related to
    validating urls.
    """

    @staticmethod
    def is_url_ok(
        url: str
    ) -> bool:
        """
        Checks if the provided url is valid. It returns True if yes or
        False if not. This method uses a head request to check the 
        status_code of the response.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        
        try:
            response = requests.head(url)

            return response.status_code == 200
        except requests.ConnectionError as e:
            # TODO: Maybe warning (?)
            #print('Something went wrong with "' + url + '" (exception raised).')
            return False
    
    def verify_image_url(
        url: str
    ):
        """
        This method will check that the provided 'url'
        is a valid one, that it is an image, and also
        will return the image file extension (if
        available), or False if not valid 'url'. The
        extension returned, if returned, contains the
        dot '.'.

        Use this method before trying to download an
        image from a url.

        This method will send a head request to the
        provided 'url', check the status_code and
        also the content-type.
        """
        # TODO: I don't like what this returns and
        # how its logic is built
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        
        try:
            response = requests.head(url)

            if response.status_code != 200: # Maybe more than only 200 are valid
                return False
            
            if not response.headers['content-type'].startswith('image/'):
                return False
            
            # This below is like 'image/jpeg', 'application/pdf' so maybe we can just
            # split by '/' and obtain the second part, because when I guess from a
            # 'image/jpeg' content-type, the response is '.jpg' and not '.jpeg'...
            content_type = response.headers['content-type']
            extension = mimetypes.guess_extension(content_type)

            # TODO: Review this below when tested 
            # This 'other_extension' below could be a better choice maybe (read above)
            other_extension = '.' + content_type.split('/')[1]

            return (
                extension
                if extension is not None else
                other_extension
            )
        except requests.ConnectionError as e:
            # TODO: Maybe warning (?)
            #print('Something went wrong with "' + url + '" (exception raised).')
            return False

    @staticmethod
    def is_youtube_video_url(
        url: str
    ):
        """
        Checks if the provided `url` is string containing
        a valid Youtube video url. This method only checks
        that the given url is well formatted.

        Example of a valid url:
        - `https://www.youtube.com/watch?v=OpA2ZxnRs6`
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)

        return GeneralRegularExpression.YOUTUBE_VIDEO_URL.parse(url)


"""
TODO: This has to use the 'yta_web_scraper' library and
also must be in another library, not this one.

I comment it to avoid errors.
"""
# from yta_general_utils.web.tiktok.url_parser import parse_tiktok_url
# from selenium import webdriver
# from selenium.webdriver.chrome.options import Options
# from selenium.webdriver.common.by import By
# from selenium.webdriver.common.action_chains import ActionChains
# from selenium.webdriver.common.keys import Keys

# import time


# def get_video_metadata_by_url(tiktok_video_url):
#     """
#     Receives a tiktok video url and build the metadata information by
#     scrapping with chromedriver.

#     This method returns the video 'id', 'username', 'video_id', 'title'
#     and 'description'
#     """
#     internal_info = parse_tiktok_url(tiktok_video_url)

#     try:
#         options = Options()
#         options.add_argument("--start-maximized")
#         # Remove this line below for debug
#         #options.add_argument("--headless=new") # for Chrome >= 109
#         driver = webdriver.Chrome(options = options)
#         driver.get(internal_info['url'])

#         time.sleep(10)

#         # [ 1 ] Continue as guest
#         actions = ActionChains(driver) 
#         actions.send_keys(Keys.TAB * 8)
#         actions.perform()

#         actions = ActionChains(driver)
#         actions.send_keys(Keys.ENTER * 1)
#         actions.perform()

#         time.sleep(5)

#         title = driver.title.replace('... | TikTok', '')
#         description = ''
#         description_elements = driver.find_elements_by_xpath("//*[@data-e2e='browse-video-desc']")[0].find_elements_by_css_selector("*")
#         for element in description_elements:
#             description += element.get_attribute('innerText') + ' '
#         description = description.strip()

#     finally:
#         driver.close()

#         return {
#             'title': title,
#             'description': description,
#             'url': internal_info['url'],
#             'username': internal_info['username'],
#             'video_id': internal_info['video_id']
#         }

# def get_popular_tiktok_videos():
#     """
#     Work in progress.
#     """
#     # TODO: In progress. To explore most popular videos and build a list of urls

#     # Create a Chrome driver
#     options = Options()
#     options.add_argument("--start-maximized")
#     driver = webdriver.Chrome(options = options)

#     try:
#         # TODO: Improve the way we look for those popular tiktok videos
#         # We can download them, but we need to be sure that those videos
#         # are actually popular and could be viral
#         """
#         driver.get(test_url)
#         time.sleep(2)
#         print(driver.current_url)
#         """

#         """
#         explore_tiktok_url = 'https://www.tiktok.com/explore'
#         driver.get(explore_tiktok_url)

#         time.sleep(5)

#         # [ 1 ] Continue as guest
#         actions = ActionChains(driver) 
#         actions.send_keys(Keys.TAB * 8)
#         actions.perform()

#         actions = ActionChains(driver)
#         actions.send_keys(Keys.ENTER * 1)
#         actions.perform()

#         time.sleep(5)

#         # [ 2 ] Look for popular videos
#         videos = driver.find_elements_by_xpath("//*[@data-e2e='explore-item']")
        
#         for video in videos:
#             # [ 3 ] Download each of those videos
#             video_link_element = video.find_elements_by_css_selector("*")[0].find_elements_by_css_selector("*")[0].find_elements_by_css_selector("*")[0]
#             href = video_link_element.get_attribute('href')
#             split = href.split('/')
#             username = str(split[len(split) - 3]).replace('@', '')
#             video_id = split[len(split) - 1]

#             # TODO: What about video title?
#             output_video_name = username + '_' + video_id + '_tiktok.mp4'
#             print('Downloading video as "' + output_video_name + '"')
#             download_tiktok_video(href, output_video_name)
#         # data-e2e="explore-item" and 3rd son is a href with video link
#         """

#         """
#         If I share a video from mobile Tiktok, the link is like this (https://vm.tiktok.com/ZGeSJ6YRA),
#         but if you search that on Chrome, it will be replaced by the whole url like this 
#         (https://www.tiktok.com/@ahorayasabesque/video/7327001175616703777?_r=1&_t=8jqq93LWqsC).
#         From that long link, if you remove the '?_...' part, you'll get the whole link that we use,
#         including the username and the video_id, so we can work propertly with it.

#         This could be a way of storing interesting tiktok videos in some places and then
#         automate the process of adding some edition to the video (maybe some short video before,
#         overlay video, title, etc.) and then upload it to youtube as a short.
#         """

#     finally:
#         driver.close()



from yta_validation.parameter import ParameterValidator
from urllib import parse as url_parse
from dataclasses import dataclass
from typing import Union


@dataclass
class UrlParameter:
    """
    Class to wrap an url parameter and make working
    with it easier.
    """

    @property
    def as_dict(
        self
    ) -> dict:
        """
        Get the UrlParameter as a dict. This method raises
        an Exception if it doesn't have a name.
        """
        if not self.can_be_dict:
            raise Exception('This UrlParameter does not have a name so it cannot be a dict.')

        return {
            self.name: self.value
        }

    @property
    def can_be_dict(
        self
    ) -> bool:
        """
        Check if this UrlParameter can be dict or not
        due to its name.
        """
        return self.name is not None
    
    @property
    def encoded(
        self
    ) -> str:
        """
        Get the parameter encoded as the RFC 3986 says
        by using the urllib library 'quote' method.

        - From 'hola dani' to 'hola%20dani'.
        - From 'key=value largo' to 'key=value%20largo'
        """
        return (
            f'{url_parse.quote(self.name)}={url_parse.quote(self.value)}'
            if self.name else
            url_parse.quote(self.value)
        )

    def __init__(
        self,
        name: Union[str, None],
        value: str
    ):
        self.name = name
        self.value = value

    @staticmethod
    def from_str(
        parameter: str
    ) -> 'UrlParameter':
        ParameterValidator.validate_mandatory_string('parameter', parameter, do_accept_empty = False)

        return UrlParameter(name = None, value = parameter)

    @staticmethod
    def from_dict(
        parameter: dict
    ) -> 'UrlParameter':
        """
        Transform the given 'parameter' dict into a UrlParameter
        instance. The provided 'parameter' parameter must be a
        dict containing only one key=value pair.
        """
        # TODO: Manipulate to make 'do_accept_empty' parameter
        ParameterValidator.validate_mandatory_dict('parameter', parameter)

        if len(parameter) != 1:
            raise Exception('The provided "parameter" does not contain only one key=value.')
        
        key, value = next(iter(parameter.items()))

        return UrlParameter(name = key, value = value)
    
    def __str__(
        self
    ) -> str:
        return (
            f'{self.name}={self.value}'
            if self.name else
            self.value
        )

@dataclass
class UrlParameters:
    """
    Class to wrap a group of UrlParameter elements to
    simplify the way we manage those items.
    """

    @property
    def as_dict(
        self
    ) -> dict:
        """
        Get the UrlParameters as a dict. This method raises
        an Exception if some Urlparameter don't have a name.
        The last keys will remain if duplicated.
        """
        if not self.can_be_dict:
            raise Exception('Some of the UrlParameters do not have a name so it cannot be a dict.')

        return {
            key: value
            for param in self.parameters
            for key, value in param.as_dict.items()
        }
    
    @property
    def can_be_dict(
        self
    ) -> bool:
        """
        Check if these UrlParameters can be dict or not
        due to their name.
        """
        return all(
            parameter.can_be_dict
            for parameter in self.parameters
        )
    
    @property
    def encoded(
        self
    ) -> str:
        """
        Get the parameters encoded as the RFC 3986 says
        by using the urllib library 'quote' method.

        - From 'hola dani&do_force' to 'hola%20dani%26do_force'.
        - From 'key=value largo&do_force' to 'key%3Dvalue%20largo%26do_forceo'
        """
        # This was previously being done with a dict
        # that included all the parameters together and
        # then 'url_parse.urlencode(parameters)', 
        # generating things like this:
        # '?query=esto%20es%20una%20frase&page=3...'.
        return '&'.join([ # %26 is & encoded
            parameter.encoded
            for parameter in self.parameters
        ])
    
    def __init__(
        self,
        parameters: list[UrlParameter]
    ):
        ParameterValidator.validate_mandatory_list_of_these_instances('parameters', parameters, UrlParameter)

        self.parameters = parameters

    @staticmethod
    def from_dict(
        parameters: dict
    ) -> 'UrlParameters':
        """
        Transform the given 'parameters' dict into a UrlParameters
        instance. The provided 'parameters' parameter must be a
        dict containing, at least, one key=value pair.
        """
        # TODO: Manipulate to make 'do_accept_empty' parameter
        ParameterValidator.validate_mandatory_dict('parameters', parameters)

        return UrlParameters([
            UrlParameter(name = key, value = value)
            for key, value in parameters.items()
        ])